/*
 * Copyright (C) 2010 SENJU Jiro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#define _GNU_SOURCE

#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <assert.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "lib/lib.h"

static void do_write(struct timespec *rts, int fd, void *p, size_t cnt,
		     off_t off)
{
	ssize_t ssz;
	struct {
		struct timespec ts;
		int err;
	} a[2];

	a[0].err = clock_gettime(CLOCK_MONOTONIC, &a[0].ts);
	ssz = pwrite(fd, p, cnt, off);
	a[1].err = clock_gettime(CLOCK_MONOTONIC, &a[1].ts);
	assert(!a[0].err);
	assert(!a[1].err);
	assert(ssz == BUFSIZ);
	if (ssz != BUFSIZ)
		abort();

	ts_sub(rts, &a[1].ts, &a[0].ts);
}

static void do_read(struct timespec *rts, int fd, void *p, size_t cnt,
		    off_t off)
{
	ssize_t ssz;
	struct {
		struct timespec ts;
		int err;
	} a[2];

	a[0].err = clock_gettime(CLOCK_MONOTONIC, &a[0].ts);
	ssz = pread(fd, p, cnt, off);
	a[1].err = clock_gettime(CLOCK_MONOTONIC, &a[1].ts);
	assert(!a[0].err);
	assert(!a[1].err);
	assert(ssz == BUFSIZ);
	if (ssz != BUFSIZ)
		abort();

	ts_sub(rts, &a[1].ts, &a[0].ts);
}

static void pr_stats(char *pre, struct timespec *ts, int n)
{
	int i;
	enum {Min, Max, Avg};
	long a[3];
	long long sum;

	a[Min] = LONG_MAX;
	a[Max] = LONG_MIN;
	sum = 0;
	for (i = 0; i < n; i++, ts++) {
		assert(!ts->tv_sec);
		if (ts->tv_nsec < a[Min])
			a[Min] = ts->tv_nsec;
		if (ts->tv_nsec > a[Max])
			a[Max] = ts->tv_nsec;
		sum += ts->tv_nsec;
	}
	a[Avg] = sum / n;

	printf("%s:\t%lu, %lu -- %ld\n", pre, a[Avg], a[Min], a[Max]);
}

int main(int argc, char *argv[])
{
	int err, fd[2], i;
	char *p;
	const int n = 10;
	struct timespec ts[n];

	fd[0] = open(argv[1], O_RDWR | O_CREAT | O_TRUNC, 0644);
	assert(fd[0] >= 0);
	fd[1] = open(argv[2], O_RDWR | O_CREAT | O_TRUNC | O_DIRECT, 0644);
	assert(fd[1] >= 0);
	err = posix_memalign((void *)&p, 512, BUFSIZ);
	assert(!err);

	for (i = 0; i < BUFSIZ; i++)
		p[i] = i % 0x0ff;

	for (i = 0; i < n; i++)
		do_write(ts + i, fd[0], p, BUFSIZ, 0);
	pr_stats("buffered write", ts, n);
	for (i = 0; i < n; i++)
		do_read(ts + i, fd[0], p, BUFSIZ, 0);
	pr_stats("buffered read", ts, n);

	for (i = 0; i < n; i++)
		do_write(ts + i, fd[1], p, BUFSIZ, 0);
	pr_stats("direct write", ts, n);
	for (i = 0; i < n; i++)
		do_read(ts + i, fd[1], p, BUFSIZ, 0);
	pr_stats("direct read", ts, n);

	return 0;
}
